from dataclasses import dataclass, field
from typing import List
from examples.voroni_world.models import Region, Villager

@dataclass
class BehaviorResults:
    villager: Villager
    revealed_regions: List[Region] = field(default_factory=list)