from dataclasses import dataclass
from behaviorTypes import BehaviorType
from typing import Dict
import cooptools.typeProviders as tp

@dataclass
class InputController:
    villager_counts: Dict[BehaviorType, tp.IntProvider]
