from cooptools.colors import Color

UI_SIDE_WIDTH_PERC = .2
UI_SIDE_WIDTH_MIN = 200
UI_SIDE_WIDTH_MAX = 300
N_REGIONS = 100
VERONI_EDGECOLOR = Color.closest_color_to_rgb((100, 100, 100))
VERONI_PERC_TO_COLOR = .3
VERONI_FILLCOLOR = Color.BLUE
NEIGHBOR_CNXN_COLOR = Color.MAROON
VILLAGE_COLOR = Color.RED
VILLAGER_MAX_ACCEL = 5
VILLAGER_MAX_VELO = 10
VILLAGER_COLOR = Color.PURPLE
MAX_VILLAGERS = 2
INIT_VILLAGERS = 1
VILLAGER_SPAWN_MS = 30000
SECTOR_TREE_DEPTH = 4
SECONDS_PER_HR = 5