from coopgame.gameTemplate import GameTemplate
import coopgame.pygamehelpers as help
from cooptools.colors import Color
from coopgame.pointdrawing import point_draw_utils as pdu

class HelpTest(GameTemplate):

    def __init__(self):

        super().__init__()

    def draw(self, frames: int, debug_mode: bool = False):
        pdu.draw_points(
            surface=self.screen,
            points={
                self.ScreenCenter: pdu.DrawPointArgs(
                    color=Color.BLUE,
                    radius=50,
                    outline_color=Color.ORANGE
                )
            }
        )

        help.draw_arrow(self.screen,
                        self.ScreenCenter,
                        help.mouse_pos_as_vector().as_tuple(),
                        Color.GREEN,
                        leader_line=True,
                        arrow_height=20)


if __name__ == "__main__":
    g = HelpTest()
    g.main()